@@ -6,8 +6,7 @@
 #  Distributed under the terms of the new BSD license.
 #
 # -----------------------------------------------------------------------------
-
-"""
+r"""
 TextBox2 provides a combination of features from TextStim and TextBox and then
 some more added:
 
@@ -17,7 +16,7 @@ some more added:
 
 """
 import numpy as np
-from pyglet import gl  # import OpenGL.GL not compatible with Big Sur (2020)
+from pyglet import gl
 
 from ..basevisual import BaseVisualStim, ColorMixin, ContainerMixin
 from psychopy.tools.attributetools import attributeSetter, setAttribute
@@ -26,7 +25,9 @@ from psychopy.tools.monitorunittools import convertToPix
 from .fontmanager import FontManager, GLFont
 from .. import shaders
 from ..rect import Rect
-from ... import core
+from ... import core, alerts
+
+from psychopy.tools.linebreak import get_breakable_points, break_units
 
 allFonts = FontManager()
 
@@ -35,6 +36,10 @@ rgbShader = None
 alphaShader = None
 showWhiteSpace = False
 
+NONE=0
+ITALIC=1
+BOLD=2
+
 codes = {'BOLD_START': u'\uE100',
          'BOLD_END': u'\uE101',
          'ITAL_START': u'\uE102',
@@ -62,6 +67,7 @@ defaultBoxWidth = {'cm': 15.0,
 
 wordBreaks = " -\n"  # what about ",."?
 
+
 END_OF_THIS_LINE = 983349843
 
 # If text is ". " we don't want to start next line with single space?
@@ -70,22 +76,21 @@ class TextBox2(BaseVisualStim, ContainerMixin, ColorMixin):
     def __init__(self, win, text, font,
                  pos=(0, 0), units=None, letterHeight=None,
                  size=None,
-                 color=(1.0, 1.0, 1.0),
-                 colorSpace='rgb',
+                 color=(1.0, 1.0, 1.0), colorSpace='rgb',
+                 fillColor=None, fillColorSpace=None,
+                 borderWidth=2, borderColor=None, borderColorSpace=None,
                  contrast=1,
-                 opacity=1.0,
+                 opacity=None,
                  bold=False,
                  italic=False,
                  lineSpacing=1.0,
                  padding=None,  # gap between box and text
                  anchor='center',
                  alignment='left',
-                 fillColor=None,
-                 borderWidth=2,
-                 borderColor=None,
                  flipHoriz=False,
                  flipVert=False,
                  editable=False,
+                 lineBreaking='default',
                  name='',
                  autoLog=None,
                  onTextCallback=None):
@@ -119,6 +124,9 @@ class TextBox2(BaseVisualStim, ContainerMixin, ColorMixin):
         flipHoriz
         flipVert
         editable
+        lineBreaking: Specifying 'default', text will be broken at a set of 
+            characters defined in the module. Specifying 'uax14', text will be 
+            broken in accordance with UAX#14 (Unicode Line Breaking Algorithm).
         name
         autoLog
         """
@@ -126,14 +134,9 @@ class TextBox2(BaseVisualStim, ContainerMixin, ColorMixin):
         BaseVisualStim.__init__(self, win, units=units, name=name)
         self.win = win
         self.colorSpace = colorSpace
-        self.color = color
-        self.contrast = contrast
-        self.opacity = opacity
+        ColorMixin.foreColor.fset(self, color)  # Have to call the superclass directly on init as text has not been set
         self.onTextCallback = onTextCallback
 
-        if units=='norm':
-            raise NotImplemented("TextBox2 doesn't support 'norm' units at the "
-                                 "moment. Use 'height' units instead")
         # first set params needed to create font (letter sizes etc)
         if letterHeight is None:
             self.letterHeight = defaultLetterHeight[self.units]
@@ -146,6 +149,9 @@ class TextBox2(BaseVisualStim, ContainerMixin, ColorMixin):
             scaleUnits = self.units
         self._pixLetterHeight = convertToPix(
                 self.letterHeight, pos=0, units=scaleUnits, win=self.win)
+        if isinstance(self._pixLetterHeight, np.ndarray):
+            # If pixLetterHeight is an array, take the Height value
+            self._pixLetterHeight = self._pixLetterHeight[1]
         self._pixelScaling = self._pixLetterHeight / self.letterHeight
         if size is None:
             size = [defaultBoxWidth[self.units], None]
@@ -158,6 +164,16 @@ class TextBox2(BaseVisualStim, ContainerMixin, ColorMixin):
         self.padding = padding
         self.glFont = None  # will be set by the self.font attribute setter
         self.font = font
+        # If font not found, default to Open Sans Regular and raise alert
+        if not self.glFont:
+            alerts.alert(4325, self, {
+                'font': font,
+                'weight': 'bold' if self.bold is True else 'regular' if self.bold is False else self.bold,
+                'style': 'italic' if self.italic else '',
+                'name': self.name})
+            self.bold = False
+            self.italic = False
+            self.font = "Open Sans"
 
         # once font is set up we can set the shader (depends on rgb/a of font)
         if self.glFont.atlas.format == 'rgb':
@@ -177,6 +193,7 @@ class TextBox2(BaseVisualStim, ContainerMixin, ColorMixin):
         # used at render time
         self._lines = None  # np.array the line numbers for each char
         self._colors = None
+        self._styles = None
         self.flipHoriz = flipHoriz
         self.flipVert = flipVert
         # params about positioning (after layout has occurred)
@@ -188,6 +205,8 @@ class TextBox2(BaseVisualStim, ContainerMixin, ColorMixin):
         self.borderWidth = borderWidth
         self.borderColor = borderColor
         self.fillColor = fillColor
+        self.contrast = contrast
+        self.opacity = opacity
 
         self.box = Rect(
                 win, pos=self.pos,
@@ -201,26 +220,50 @@ class TextBox2(BaseVisualStim, ContainerMixin, ColorMixin):
                 units=self.units,
                 lineWidth=1, lineColor=None, fillColor=fillColor, opacity=0.1,
                 autoLog=False)
-        self.pallette = { # If no focus
-                'lineColor': borderColor,
-                'lineWidth': borderWidth,
-                'fillColor': fillColor,
-        }
+        # set linebraking option
+        if lineBreaking not in ('default', 'uax14'):
+            raise ValueError("Unknown lineBreaking option ({}) is"
+                "specified.".format(lineBreaking))
+        self._lineBreaking = lineBreaking
         # then layout the text (setting text triggers _layout())
-        self.startText = text
-        self.text = text if text is not None else ""
+        self._text = ''
+        self.text = self.startText = text if text is not None else ""
 
         # caret
-        self.editable = editable
+        self._editable = editable
         self.caret = Caret(self, color=self.color, width=5)
-        self._hasFocus = False
-        if editable:  # may yet gain focus if the first editable obj
-            self.win.addEditable(self)
+
 
         self.autoLog = autoLog
 
+    @property
+    def editable(self):
+        return self._editable
+    
+    @editable.setter
+    def editable(self, editable):
+        self._editable = editable
+        if editable is False and self.hasFocus:
+            if self.win:
+                self.win.removeEditable(self)
+        if editable is True:
+            if self.win:
+                self.win.addEditable(self)
+        
     @property
     def pallette(self):
+        self._pallette = {
+            False: {
+                'lineColor': self._borderColor,
+                'lineWidth': self.borderWidth,
+                'fillColor': self._fillColor
+            },
+            True: {
+                'lineColor': self._borderColor-0.1,
+                'lineWidth': self.borderWidth+1,
+                'fillColor': self._fillColor+0.1
+            }
+        }
         return self._pallette[self.hasFocus]
 
     @pallette.setter
@@ -230,6 +273,16 @@ class TextBox2(BaseVisualStim, ContainerMixin, ColorMixin):
             True: value
         }
 
+    @property
+    def foreColor(self):
+        return ColorMixin.foreColor.fget(self)
+    @foreColor.setter
+    def foreColor(self, value):
+        ColorMixin.foreColor.fset(self, value)
+        self._layout()
+        if hasattr(self, "foreColor") and hasattr(self, 'caret'):
+            self.caret.color = self._foreColor
+
     @attributeSetter
     def font(self, fontName, italic=False, bold=False):
         if isinstance(fontName, GLFont):
@@ -242,6 +295,17 @@ class TextBox2(BaseVisualStim, ContainerMixin, ColorMixin):
                     size=int(round(self._pixLetterHeight)),
                     bold=self.bold, italic=self.italic)
 
+    @property
+    def fontMGR(self):
+            return allFonts
+    @fontMGR.setter
+    def fontMGR(self, mgr):
+        global allFonts
+        if isinstance(mgr, FontManager):
+            allFonts = mgr
+        else:
+            raise TypeError(f"Could not set font manager for TextBox2 object `{self.name}`, must be supplied with a FontManager object")
+
     @attributeSetter
     def anchor(self, anchor):
         """anchor is a string of terms, top, bottom, left, right, center
@@ -287,36 +351,104 @@ class TextBox2(BaseVisualStim, ContainerMixin, ColorMixin):
 
         self._needVertexUpdate = True
 
-    @attributeSetter
+    @property
+    def text(self):
+        lastFormatter = NONE
+        formatted_text = ''
+        styles = self._styles
+        for i, c in enumerate(self._text):
+            if styles[i] == ITALIC and lastFormatter != styles[i]:
+                formatted_text+='<i>%s'%(c)
+            elif styles[i] == BOLD and lastFormatter != styles[i]:
+                formatted_text+='<b>%s'%(c)
+            
+            elif styles[i] != ITALIC and lastFormatter == ITALIC:
+                formatted_text+='</i>%s'%(c)
+            elif styles[i] != BOLD and lastFormatter == BOLD:
+                formatted_text+='</b>%s'%(c)
+            else:
+                formatted_text+=c
+            lastFormatter = styles[i]
+        return formatted_text
+    
+    @text.setter
     def text(self, text):
-        self.__dict__['text'] = text
+        text = text.replace('<i>', codes['ITAL_START'])
+        text = text.replace('</i>', codes['ITAL_END'])
+        text = text.replace('<b>', codes['BOLD_START'])
+        text = text.replace('</b>', codes['BOLD_END'])      
+        visible_text = ''.join([c for c in text if c not in codes.values()])
+        self._styles = [0,]*len(visible_text)
+        self._text = visible_text
+        
+        current_style=0
+        ci = 0
+        for c in text:
+            if c == codes['ITAL_START']:
+                current_style += ITALIC
+            elif c == codes['BOLD_START']:
+                current_style += BOLD
+            elif c == codes['BOLD_END']:
+                current_style -= BOLD
+            elif c == codes['ITAL_END']:
+                current_style -= ITALIC
+            else:
+                self._styles[ci]=current_style
+                ci+=1
+                
         self._layout()
 
+    def addCharAtCaret(self, char):
+        txt = self._text
+        txt = txt[:self.caret.index] + char + txt[self.caret.index:]
+        cstyle = NONE
+        if len(self._styles) and self.caret.index <= len(self._styles):
+            cstyle = self._styles[self.caret.index-1]
+        self._styles.insert(self.caret.index, cstyle)
+        self.caret.index += 1
+        self._text = txt
+        self._layout()
+
+    def deleteCaretLeft(self):
+        if self.caret.index > 0:
+            txt = self._text
+            ci = self.caret.index
+            txt = txt[:ci-1] + txt[ci:]
+            self._styles = self._styles[:ci-1]+self._styles[ci:]
+            self.caret.index -= 1
+            self._text = txt
+            self._layout()
+
+    def deleteCaretRight(self):
+        ci = self.caret.index
+        if ci < len(self._text):
+            txt = self._text
+            txt = txt[:ci] + txt[ci+1:]
+            self._styles = self._styles[:ci]+self._styles[ci+1:]
+            self._text = txt
+            self._layout()
+        
     def _layout(self):
         """Layout the text, calculating the vertex locations
         """
         def getLineWidthFromPix(pixVal):
             return pixVal / self._pixelScaling + self.padding * 2
-
-        text = self.text
-        text = text.replace('<i>', codes['ITAL_START'])
-        text = text.replace('</i>', codes['ITAL_END'])
-        text = text.replace('<b>', codes['BOLD_START'])
-        text = text.replace('</b>', codes['BOLD_END'])
-        rgb = self._getDesiredRGB(self.rgb, self.colorSpace, self.contrast)
+        
+        rgb = self._foreColor.render('rgba1')
         font = self.glFont
 
         # the vertices are initially pix (natural for freetype)
         # then we convert them to the requested units for self._vertices
         # then they are converted back during rendering using standard BaseStim
-        vertices = np.zeros((len(text) * 4, 2), dtype=np.float32)
-        self._charIndices = np.zeros((len(text)), dtype=int)
-        self._colors = np.zeros((len(text) * 4, 4), dtype=np.float32)
-        self._texcoords = np.zeros((len(text) * 4, 2), dtype=np.float32)
-        self._glIndices = np.zeros((len(text) * 4), dtype=int)
+        visible_text = self._text
+        vertices = np.zeros((len(visible_text) * 4, 2), dtype=np.float32)
+        self._charIndices = np.zeros((len(visible_text)), dtype=int)
+        self._colors = np.zeros((len(visible_text) * 4, 4), dtype=np.double)
+        self._texcoords = np.zeros((len(visible_text) * 4, 2), dtype=np.double)
+        self._glIndices = np.zeros((len(visible_text) * 4), dtype=int)
 
         # the following are used internally for layout
-        self._lineNs = np.zeros(len(text), dtype=int)
+        self._lineNs = np.zeros(len(visible_text), dtype=int)
         self._lineTops = []  # just length of nLines
         self._lineBottoms = []
         self._lineLenChars = []  #
@@ -338,120 +470,272 @@ class TextBox2(BaseVisualStim, ContainerMixin, ColorMixin):
         else:
             alphaCorrection = 1
 
-        wordLen = 0
-        charsThisLine = 0
-        wordsThisLine = 0
-        lineN = 0
+        if self._lineBreaking == 'default':
 
-        for i, charcode in enumerate(text):
+            wordLen = 0
+            charsThisLine = 0
+            wordsThisLine = 0
+            lineN = 0
 
-            printable = True  # unless we decide otherwise
-            # handle formatting codes
-            if charcode in codes.values():
-                if charcode == codes['ITAL_START']:
-                    fakeItalic = 0.1 * font.size
-                elif charcode == codes['ITAL_END']:
+            for i, charcode in enumerate(self._text):
+                printable = True  # unless we decide otherwise
+                # handle formatting codes
+                if self._styles[i] == NONE:
                     fakeItalic = 0.0
-                elif charcode == codes['BOLD_START']:
-                    fakeBold = 0.3 * font.size
-                elif charcode == codes['BOLD_END']:
-                    current[0] -= fakeBold / 2  # we expected bigger current
                     fakeBold = 0.0
-                continue
-            # handle newline
-            if charcode == '\n':
-                printable = False
+                elif self._styles[i] == ITALIC:
+                    fakeItalic = 0.1 * font.size
+                elif self._styles[i] == ITALIC:
+                    fakeBold = 0.3 * font.size
 
-            # handle printable characters
-            if printable:
-                if showWhiteSpace and charcode == " ":
-                    glyph = font[u"·"]
-                else:
+                # handle newline
+                if charcode == '\n':
+                    printable = False
+
+                # handle printable characters
+                if printable:
                     glyph = font[charcode]
-                xBotL = current[0] + glyph.offset[0] - fakeItalic - fakeBold / 2
-                xTopL = current[0] + glyph.offset[0] - fakeBold / 2
-                yTop = current[1] + glyph.offset[1]
-                xBotR = xBotL + glyph.size[0] * alphaCorrection + fakeBold
-                xTopR = xTopL + glyph.size[0] * alphaCorrection + fakeBold
-                yBot = yTop - glyph.size[1]
-                u0 = glyph.texcoords[0]
-                v0 = glyph.texcoords[1]
-                u1 = glyph.texcoords[2]
-                v1 = glyph.texcoords[3]
-            else:
-                glyph = font[u"·"]
-                x = current[0] + glyph.offset[0]
-                yTop = current[1] + glyph.offset[1]
-                yBot = yTop - glyph.size[1]
-                xBotL = x
-                xTopL = x
-                xBotR = x
-                xTopR = x
-                u0 = glyph.texcoords[0]
-                v0 = glyph.texcoords[1]
-                u1 = glyph.texcoords[2]
-                v1 = glyph.texcoords[3]
+                    if showWhiteSpace and charcode == " ":
+                        glyph = font[u"·"]
+                    elif charcode == " ":
+                        # glyph size of space is smaller than actual size, so use size of dot instead
+                        glyph.size = font[u"·"].size
+                    xBotL = current[0] + glyph.offset[0] - fakeItalic - fakeBold / 2
+                    xTopL = current[0] + glyph.offset[0] - fakeBold / 2
+                    yTop = current[1] + glyph.offset[1]
+                    xBotR = xBotL + glyph.size[0] * alphaCorrection + fakeBold
+                    xTopR = xTopL + glyph.size[0] * alphaCorrection + fakeBold
+                    yBot = yTop - glyph.size[1]
+                    # Adjust for norm
+                    if self.units == 'norm':
+                        ratio = self.win.size[1]/self.win.size[0]
+                        xBotL *= ratio
+                        xTopL *= ratio
+                        xBotR *= ratio
+                        xTopR *= ratio
+                    u0 = glyph.texcoords[0]
+                    v0 = glyph.texcoords[1]
+                    u1 = glyph.texcoords[2]
+                    v1 = glyph.texcoords[3]
+                else:
+                    glyph = font[u"·"]
+                    x = current[0] + glyph.offset[0]
+                    yTop = current[1] + glyph.offset[1]
+                    yBot = yTop - glyph.size[1]
+                    xBotL = x
+                    xTopL = x
+                    xBotR = x
+                    xTopR = x
+                    u0 = glyph.texcoords[0]
+                    v0 = glyph.texcoords[1]
+                    u1 = glyph.texcoords[2]
+                    v1 = glyph.texcoords[3]
 
-            index = i * 4
-            theseVertices = [[xTopL, yTop], [xBotL, yBot],
-                             [xBotR, yBot], [xTopR, yTop]]
-            texcoords = [[u0, v0], [u0, v1],
-                         [u1, v1], [u1, v0]]
+                theseVertices = [[xTopL, yTop], [xBotL, yBot],
+                                 [xBotR, yBot], [xTopR, yTop]]
+                texcoords = [[u0, v0], [u0, v1],
+                             [u1, v1], [u1, v0]]
 
-            vertices[i * 4:i * 4 + 4] = theseVertices
-            self._texcoords[i * 4:i * 4 + 4] = texcoords
-            self._colors[i*4 : i*4+4, :3] = rgb
-            self._colors[i*4 : i*4+4, 3] = self.opacity
-            self._lineNs[i] = lineN
-            current[0] = current[0] + glyph.advance[0] + fakeBold / 2
-            current[1] = current[1] + glyph.advance[1]
+                vertices[i * 4:i * 4 + 4] = theseVertices
+                self._texcoords[i * 4:i * 4 + 4] = texcoords
+                self._colors[i*4 : i*4+4, :4] = rgb
+                self._lineNs[i] = lineN
+                current[0] = current[0] + glyph.advance[0] + fakeBold / 2
+                current[1] = current[1] + glyph.advance[1]
 
-            # are we wrapping the line?
-            if charcode == "\n":
-                lineWPix = current[0]
-                current[0] = 0
-                current[1] -= self._lineHeight
-                lineN += 1
-                charsThisLine += 1
-                self._lineLenChars.append(charsThisLine)
-                self._lineWidths.append(getLineWidthFromPix(lineWPix))
-                charsThisLine = 0
-                wordsThisLine = 0
-            elif charcode in wordBreaks:
-                wordLen = 0
-                charsThisLine += 1
-                wordsThisLine += 1
-            elif printable:
-                wordLen += 1
-                charsThisLine += 1
+                # are we wrapping the line?
+                if charcode == "\n":
+                    lineWPix = current[0]
+                    current[0] = 0
+                    current[1] -= self._lineHeight
+                    lineN += 1
+                    charsThisLine += 1
+                    self._lineLenChars.append(charsThisLine)
+                    self._lineWidths.append(getLineWidthFromPix(lineWPix))
+                    charsThisLine = 0
+                    wordsThisLine = 0
+                elif charcode in wordBreaks:
+                    wordLen = 0
+                    charsThisLine += 1
+                    wordsThisLine += 1
+                elif printable:
+                    wordLen += 1
+                    charsThisLine += 1
 
-            # end line with auto-wrap on space
-            if current[0] >= lineMax and wordLen > 0 and wordsThisLine:
-                # move the current word to next line
-                lineBreakPt = vertices[(i - wordLen + 1) * 4, 0]
-                wordWidth = current[0] - lineBreakPt
-                # shift all chars of the word left by wordStartX
-                vertices[(i - wordLen + 1) * 4: (i + 1) * 4, 0] -= lineBreakPt
-                vertices[(i - wordLen + 1) * 4: (i + 1) * 4, 1] -= self._lineHeight
-                # update line values
-                self._lineNs[i - wordLen + 1: i + 1] += 1
-                self._lineLenChars.append(charsThisLine - wordLen)
-                self._lineWidths.append(getLineWidthFromPix(lineBreakPt))
-                lineN += 1
-                # and set current to correct location
-                current[0] = wordWidth
-                current[1] -= self._lineHeight
-                charsThisLine = wordLen
+                # end line with auto-wrap on space
+                if current[0] >= lineMax and wordLen > 0 and wordsThisLine > 1:
+                    # move the current word to next line
+                    lineBreakPt = vertices[(i - wordLen + 1) * 4, 0]
+                    wordWidth = current[0] - lineBreakPt
+                    # shift all chars of the word left by wordStartX
+                    vertices[(i - wordLen + 1) * 4: (i + 1) * 4, 0] -= lineBreakPt
+                    vertices[(i - wordLen + 1) * 4: (i + 1) * 4, 1] -= self._lineHeight
+                    # update line values
+                    self._lineNs[i - wordLen + 1: i + 1] += 1
+                    self._lineLenChars.append(charsThisLine - wordLen)
+                    self._lineWidths.append(getLineWidthFromPix(lineBreakPt))
+                    lineN += 1
+                    # and set current to correct location
+                    current[0] = wordWidth
+                    current[1] -= self._lineHeight
+                    charsThisLine = wordLen
+                    wordsThisLine = 1
 
-            # have we stored the top/bottom of this line yet
-            if lineN + 1 > len(self._lineTops):
-                self._lineBottoms.append(current[1] + font.descender)
-                self._lineTops.append(current[1] + self._lineHeight
-                                      + font.descender/2)
+                # have we stored the top/bottom of this line yet
+                if lineN + 1 > len(self._lineTops):
+                    self._lineBottoms.append(current[1] + font.descender)
+                    self._lineTops.append(current[1] + self._lineHeight
+                                          + font.descender/2)
+                
+            # finally add length of this (unfinished) line
+            self._lineWidths.append(getLineWidthFromPix(current[0]))
+            self._lineLenChars.append(charsThisLine)
 
-        # finally add length of this (unfinished) line
-        self._lineWidths.append(getLineWidthFromPix(current[0]))
-        self._lineLenChars.append(charsThisLine)
+        elif self._lineBreaking == 'uax14':
+
+            # get a list of line-breakable points according to UAX#14
+            breakable_points = list(get_breakable_points(self._text))
+            text_seg = list(break_units(self._text, breakable_points))
+            styles_seg = list(break_units(self._styles, breakable_points))
+
+            lineN = 0
+            charwidth_list = []
+            segwidth_list = []
+            y_advance_list = []
+            vertices_list = []
+            texcoords_list = []
+
+            # calculate width of each segments
+            for this_seg in range(len(text_seg)):
+
+                thisSegWidth = 0 # width of this segment
+
+                for i, charcode in enumerate(text_seg[this_seg]):
+                    printable = True  # unless we decide otherwise
+                    # handle formatting codes
+                    if styles_seg[this_seg][i] == NONE:
+                        fakeItalic = 0.0
+                        fakeBold = 0.0
+                    elif styles_seg[this_seg][i] == ITALIC:
+                        fakeItalic = 0.1 * font.size
+                    elif styles_seg[this_seg][i] == ITALIC:
+                        fakeBold = 0.3 * font.size
+
+                    # handle newline
+                    if charcode == '\n':
+                        printable = False
+
+                    # handle printable characters
+                    if printable:
+                        if showWhiteSpace and charcode == " ":
+                            glyph = font[u"·"]
+                        else:
+                            glyph = font[charcode]
+                        xBotL = glyph.offset[0] - fakeItalic - fakeBold / 2
+                        xTopL = glyph.offset[0] - fakeBold / 2
+                        yTop = glyph.offset[1]
+                        xBotR = xBotL + glyph.size[0] * alphaCorrection + fakeBold
+                        xTopR = xTopL + glyph.size[0] * alphaCorrection + fakeBold
+                        yBot = yTop - glyph.size[1]
+                        u0 = glyph.texcoords[0]
+                        v0 = glyph.texcoords[1]
+                        u1 = glyph.texcoords[2]
+                        v1 = glyph.texcoords[3]
+                    else:
+                        glyph = font[u"·"]
+                        x = glyph.offset[0]
+                        yTop = glyph.offset[1]
+                        yBot = yTop - glyph.size[1]
+                        xBotL = x
+                        xTopL = x
+                        xBotR = x
+                        xTopR = x
+                        u0 = glyph.texcoords[0]
+                        v0 = glyph.texcoords[1]
+                        u1 = glyph.texcoords[2]
+                        v1 = glyph.texcoords[3]
+
+                    # calculate width and update segment width
+                    w = glyph.advance[0] + fakeBold / 2
+                    thisSegWidth += w
+
+                    # keep vertices, texcoords, width and y_advance of this character
+                    vertices_list.append([[xTopL, yTop], [xBotL, yBot],
+                                          [xBotR, yBot], [xTopR, yTop]])
+                    texcoords_list.append([[u0, v0], [u0, v1],
+                                           [u1, v1], [u1, v0]])
+                    charwidth_list.append(w)
+                    y_advance_list.append(glyph.advance[1])
+
+                # append width of this segment to the list
+                segwidth_list.append(thisSegWidth)
+
+            # concatenate segments to build line
+            lines = []
+            while text_seg:
+                line_width = 0
+                for i in range(len(text_seg)):
+                    # if this segment is \n, break line here.
+                    if text_seg[i][-1] == '\n':
+                        i+=1 # increment index to include \n to current line
+                        break
+                    # concatenate next segment
+                    line_width += segwidth_list[i]
+                    # break if line_width is greater than lineMax
+                    if lineMax < line_width:
+                        break
+                else:
+                    # if for sentence finished without break, all segments 
+                    # should be concatenated.
+                    i = len(text_seg)
+                p = max(1, i)
+                # concatenate segments and remove from segment list
+                lines.append("".join(text_seg[:p]))
+                del text_seg[:p], segwidth_list[:p] #, avoid[:p]
+
+            # build lines
+            i = 0 # index of the current character
+            if lines:
+                for line in lines:
+                    for c in line:
+                        theseVertices = vertices_list[i]
+                        #update vertices
+                        for j in range(4):
+                            theseVertices[j][0] += current[0]
+                            theseVertices[j][1] += current[1]
+                        texcoords = texcoords_list[i]
+
+                        vertices[i * 4:i * 4 + 4] = theseVertices
+                        self._texcoords[i * 4:i * 4 + 4] = texcoords
+                        self._colors[i*4 : i*4+4, :4] = rgb
+                        self._lineNs[i] = lineN
+
+                        current[0] = current[0] + charwidth_list[i]
+                        current[1] = current[1] + y_advance_list[i]
+                        
+                        # have we stored the top/bottom of this line yet
+                        if lineN + 1 > len(self._lineTops):
+                            self._lineBottoms.append(current[1] + font.descender)
+                            self._lineTops.append(current[1] + self._lineHeight
+                                                   + font.descender/2)
+
+                        # next chacactor
+                        i += 1
+
+                    # prepare for next line
+                    current[0] = 0
+                    current[1] -= self._lineHeight
+                    
+                    lineBreakPt = vertices[(i-1) * 4, 0]
+                    self._lineLenChars.append(len(line))
+                    self._lineWidths.append(getLineWidthFromPix(lineBreakPt))
+
+                    # need not increase lineN when the last line doesn't end with '\n'
+                    if lineN < len(lines)-1 or line[-1] == '\n' :
+                        lineN += 1
+        else:
+            raise ValueError("Unknown lineBreaking option ({}) is"
+                "specified.".format(lineBreaking))
 
         # convert the vertices to stimulus units
         self._rawVerts = vertices / self._pixelScaling
@@ -492,6 +776,7 @@ class TextBox2(BaseVisualStim, ContainerMixin, ColorMixin):
         #self.fillColor = self.box.fillColor
 
         if self._needVertexUpdate:
+            #print("Updating vertices...")
             self._updateVertices()
         if self.fillColor is not None or self.borderColor is not None:
             self.box.draw()
@@ -510,16 +795,16 @@ class TextBox2(BaseVisualStim, ContainerMixin, ColorMixin):
         gl.glEnableClientState(gl.GL_TEXTURE_COORD_ARRAY)
         gl.glEnableClientState(gl.GL_VERTEX_ARRAY)
 
-        gl.glVertexPointer(2, gl.GL_FLOAT, 0, self.verticesPix)
-        gl.glColorPointer(4, gl.GL_FLOAT, 0, self._colors)
-        gl.glTexCoordPointer(2, gl.GL_FLOAT, 0, self._texcoords)
+        gl.glVertexPointer(2, gl.GL_DOUBLE, 0, self.verticesPix.ctypes)
+        gl.glColorPointer(4, gl.GL_DOUBLE, 0, self._colors.ctypes)
+        gl.glTexCoordPointer(2, gl.GL_DOUBLE, 0, self._texcoords.ctypes)
 
         self.shader.bind()
         self.shader.setInt('texture', 0)
         self.shader.setFloat('pixel', [1.0 / 512, 1.0 / 512])
-        nVerts = len(self.text)*4
-        gl.glDrawElements(gl.GL_QUADS, nVerts,
-                          gl.GL_UNSIGNED_INT, list(range(nVerts)))
+        nVerts = len(self._text)*4
+
+        gl.glDrawArrays(gl.GL_QUADS, 0, nVerts)
         self.shader.unbind()
 
         # removed the colors and font texture
@@ -540,9 +825,10 @@ class TextBox2(BaseVisualStim, ContainerMixin, ColorMixin):
     def reset(self):
         # Reset contents
         self.text = self.startText
-        # Make sure box is still editable (if needed)
-        if self.editable and self not in self.win._editableChildren:  # may yet gain focus if the first editable obj
-            self.win.addEditable(self)
+
+    def clear(self):
+        # Clear contents
+        self.text = ""
 
 
     def contains(self, x, y=None, units=None, tight=False):
@@ -677,9 +963,7 @@ class TextBox2(BaseVisualStim, ContainerMixin, ColorMixin):
             return
         if chr == '\r':  # make it newline not Carriage Return
             chr = '\n'
-        txt = self.text
-        self.text = txt[:self.caret.index] + chr + txt[self.caret.index:]
-        self.caret.index += 1
+        self.addCharAtCaret(chr)
         if self.onTextCallback:
             self.onTextCallback()
 
@@ -694,10 +978,9 @@ class TextBox2(BaseVisualStim, ContainerMixin, ColorMixin):
         elif key == 'MOTION_LEFT':
             self.caret.char -= 1
         elif key == 'MOTION_BACKSPACE':
-            self.text = self.text[:self.caret.index-1] + self.text[self.caret.index:]
-            self.caret.index -= 1
+            self.deleteCaretLeft()
         elif key == 'MOTION_DELETE':
-            self.text = self.text[:self.caret.index] + self.text[self.caret.index+1:]
+            self.deleteCaretRight()
         elif key == 'MOTION_NEXT_WORD':
             pass
         elif key == 'MOTION_PREVIOUS_WORD':
@@ -719,19 +1002,37 @@ class TextBox2(BaseVisualStim, ContainerMixin, ColorMixin):
 
     @property
     def hasFocus(self):
-        return self._hasFocus
+        if self.win and self.win.currentEditable == self:
+            return True
+        return False
 
     @hasFocus.setter
-    def hasFocus(self, state):
-        # Store focus
-        self._hasFocus = state
-        # Redraw text box
-        self.draw()
+    def hasFocus(self, focus):
+        if focus is False and self.hasFocus:
+            # If focus is being set to False, tell window to 
+            # give focus to next editable.
+            if self.win:
+                self.win.nextEditable()
+        elif focus is True and self.hasFocus is False:
+            # If focus is being set True, set textbox instance to be
+            # window.currentEditable.
+            if self.win:
+                self.win.currentEditable=self
+        return False
 
     def getText(self):
-        """Returns the current text in the box"""
+        """Returns the current text in the box, including formating tokens."""
         return self.text
 
+    @property
+    def visibleText(self):
+        """Returns the current visible text in the box"""
+        return self._text
+
+    def getVisibleText(self):
+        """Returns the current visible text in the box"""
+        return self.visibleText
+
     @attributeSetter
     def pos(self, value):
         """The position of the center of the TextBox in the stimulus
@@ -806,31 +1107,21 @@ class Caret(ColorMixin):
 
     def __init__(self, textbox, color, width, colorSpace='rgb'):
         self.textbox = textbox
-        self.index = len(textbox.text)  # start off at the end
+        self.index = len(textbox._text)  # start off at the end
         self.autoLog = False
         self.width = width
         self.units = textbox.units
         self.colorSpace = colorSpace
         self.color = color
 
-    @attributeSetter
-    def color(self, color):
-        self.setColor(color)
-        self._desiredRGB = [0.89, -0.35, -0.28]
-        # if self.colorSpace not in ['rgb', 'dkl', 'lms', 'hsv']:
-        #     self._desiredRGB = [c / 127.5 - 1 for c in self.rgb]
-        # else:
-        #     self._desiredRGB = self.rgb
-
     def draw(self):
         if not self.visible:
             return
         if core.getTime() % 1 > 0.6:  # Flash every other second
             return
         gl.glLineWidth(self.width)
-        rgb = self._desiredRGB
         gl.glColor4f(
-            rgb[0], rgb[1], rgb[2], self.textbox.opacity
+            *self._foreColor.rgba1
         )
         gl.glBegin(gl.GL_LINES)
         gl.glVertex2f(self.vertices[0, 0], self.vertices[0, 1])
@@ -917,8 +1208,8 @@ class Caret(ColorMixin):
     def vertices(self):
         textbox = self.textbox
         # check we have a caret index
-        if self.index is None or self.index > len(textbox.text):
-            self.index = len(textbox.text)
+        if self.index is None or self.index > len(textbox._text):
+            self.index = len(textbox._text)
         if self.index < 0:
             self.index = 0
         # get the verts of character next to caret (chr is the next one so use
